/*Copyright (c) 2012 Moussa Dembélé (www.mousman.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package com.mousman.as3.agal.parser.objs 
{
	import com.greensock.TweenLite;
	import com.mousman.as3.agal.parser.dto.MBuffersWriterDTO;
	import com.mousman.as3.structure.garbage.IDestroyable;
	import flash.events.Event;
	import flash.events.EventDispatcher;
	import flash.utils.ByteArray;
	import org.osflash.signals.Signal;
	/**
	 * creates a set of vertex buffers and index buffers based on a .obj and .mtl file
	 * and exports them in a binary file
	 * @author Moussa Dembélé (www.mousman.com) based on Christer Kaitila and Nathan Ostgard work
	 * 
	 */
	public class MCompactObjExportParser extends EventDispatcher implements IDestroyable
	{
		// constants used in parsing OBJ data
		private const LINE_FEED:String = String.fromCharCode(10);
		private const SPACE:String = String.fromCharCode(32);
		private const SLASH:String = "/";
		private const VERTEX:String = "v";
		private const NORMAL:String = "vn";
		private const UV:String = "vt";
		private const INDEX_DATA:String = "f";	
		private const MTL_FILE:String = "mtllib";	
		private const USE_MTL:String = "usemtl";	
		private const OBJECT_NAME:String = "o";	
		private const GROUP_NAME:String = "g";	
		private const SMOOTH_SHADING:String = "s";
		
		private var _mtlFile:String;		
		// older versions of 3dsmax use an invalid vertex order:
		private var _vertexDataIsZxy:Boolean = false;
		private var _scale:Number;
		private var _writerParams:MBuffersWriterDTO 

		
		// the parser for the mtl datas
		private var _mtlParser:MMtlParser;
		// contains all the mtl datas parsed for the mtl file
		private var _mtlSets:Vector.<MMtlSet>;
		
		private var _currentObject:String;
		private var _currentGroup:String;
		// contains all the polygons group for the current object
		
		private var _defaultMtlSet:MMtlSet;
		private var _currentMtlSet:MMtlSet;
		
		// some exporters mirror the UV texture coordinates
		private var _mirrorUV:Boolean = true;
		
		private var _lineSignal:Signal;
		private var _signal:Signal;
		
		private var _lines:Array;
		private var _currentLineIndex:int;
		private const MAX_POLYGONES_PER_GROUP:int = 21248;
		private var _currentPolygoneNumber:int = 0;		
		private var _totalPolygoneNumber:int = 0;		
		
		private var _exportBA:ByteArray;
		private var _onVertex:Boolean = false;
		private var _onNormal:Boolean= false;
		private var _onUV:Boolean = false;
		private var _onFaces:Boolean = false;
		private var _verticesNumber:uint=0;
		private var _normalesNumber:uint=0;
		private var _uvNumber:uint=0;
		private var _facesNumber:uint=0;
		private var _verticesDataPos:uint;
		private var _normalesDataPos:uint;
		private var _uvDataPos:uint;
		private var _facesDataPos:uint;
		
		private var _smoothPos:uint;
		private var _smoothValue:Number;
		
		
		/**
		* Constructor. Creates a new MObjParser instance.
		*
		* <p>After you create the MObjParser instance, call the parse method to create the mesh datas. And then the export method</p>
		* 
		* @param writerStore : WriterStore.<p>Used to resize the object</p>
		* @param dataIsZxy : Boolean.<p>Older versions of 3dsmax use an invalid vertex order</p>
		* @param mirrorUV : Boolean.<p>Some exporters mirror the UV texture coordinates</p>
		*/		
		public function MCompactObjExportParser(dataIsZxy:Boolean = false, mirrorUV:Boolean = true) {
			// older versions of 3dsmax use an invalid vertex order:
			_vertexDataIsZxy = dataIsZxy;
			// some exporters mirror the UV texture coordinates
			_mirrorUV = mirrorUV;
			//_writerStore = writerStore;
		}
		//----------------------------------------------------------------------------------------------
		/**
		* Parse the obj and mtl file and create the mesh datas
		*
		* 
		* @param objfile : String.<p>Contains the obj file datas</p>
		* @param mtlfile : String.<p>Contains the mtl file datas</p>
		* @param scale : Number.<p>Used to resize the 3D model</p>
		* @param writerParams : used to force color on the mesh.<p></p>
		* 
		* @return the datas of the mesh
		*/			
		//----------------------------------------------------------------------------------------------		
		public function parse(objfile:String, mtlfile:String = null, scale:Number=1, writerParams:MBuffersWriterDTO = null): void {
			_scale = scale;
			TweenLite.delayedCall(1,startParsing,[objfile,mtlfile,writerParams]);
			
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function startParsing(objfile:String, mtlfile:String = null, writerParams:MBuffersWriterDTO = null):void 
		{
			if(writerParams == null) 
			_writerParams = new MBuffersWriterDTO();
			// read and explode the mtl datas
			if(mtlfile)
				parseMtl(mtlfile);
			_defaultMtlSet = new MMtlSet("default");
			_defaultMtlSet.ambiantColor = 0xcccccc;
			_defaultMtlSet.diffuseColor = 0xcccccc;
			_defaultMtlSet.dissolved= 1;
			_defaultMtlSet.emissiveColor = 0x0;
			_defaultMtlSet.specularCoef = 100;
			_defaultMtlSet.specularColor= 0xffffff;
			// read and explode the obj datas
			
			_exportBA = new ByteArray();
			writeHeader();
			
			parseLines(objfile);
		}
		//----------------------------------------------------------------------------------------------
		// parse the mtl file and create a set of MMtlSet
		//----------------------------------------------------------------------------------------------
		private function parseMtl(mtlfile:String):void {
			_mtlParser = new MMtlParser(mtlfile);
			_mtlParser.parse(_writerParams);
			_mtlSets = _mtlParser.mtlSetVector;
		}	
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function parseLines(objfile:String):void {
			_lines = objfile.split(LINE_FEED);
			
			var line:String;
			var loop:uint = _lines.length;		
			var id:String;
			
			if (_lineSignal) _lineSignal.remove(parseBlock);
			_lineSignal = new Signal();
			_lineSignal.add(parseBlock);
			_currentLineIndex = 0;
			parseBlock();
			
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function parseBlock():void 
		{
			log("parseBlock");
			nextBlock();
		}	
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function nextBlock():void 
		{
			var line:String;
			var out:Boolean = false;
		
			var loop:int = 500;
			var max:int = _lines.length;
			while(!out && _currentLineIndex <= max) {
				// last line
				if (_currentLineIndex == max) {
					// complete last polygone group
					log("complete");
					completePolygonGroup();
					dispatchEvent(new Event(Event.COMPLETE));
					return;
				}
				line = _lines[_currentLineIndex];
				line.replace(/^\s*|\s*$/g, '');
				parseLine(line);
				_currentLineIndex ++;
				if (_currentPolygoneNumber == MAX_POLYGONES_PER_GROUP) {
					completePolygonGroup();
					// we create a group for the next bloc of polygones and assign smooth shading and the mtl data of the current group
					var sValue:Number = _smoothValue;
					setNewGroup("group_" + _currentLineIndex.toString());
					
					_exportBA.position = _smoothPos;
					_exportBA.writeDouble(sValue);	
					_smoothValue = sValue;
					
					writeMtl(_currentMtlSet);
					
					out = true;
				}
			}
			
			if (out) {
				callNext();
			}
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		public function callNext():void {
			_lineSignal.dispatch();
		}		
		//----------------------------------------------------------------------------------------------	
		//----------------------------------------------------------------------------------------------		
		private function parseLine(line:String):void {
			var name:String;
			// Split line into words.
			var words:Array = line.split(SPACE);
			// Blank line or comment, ignore it
			if (words.length ==0 || line === '' || line.charAt(0) === '#') {
				return;
			}
			// Prepare the data of the line.
			var data:Array = words.slice(1);
			
			// Check first word and delegate remainder to proper parser.
			while ((data[0] == '') || (data[0] == ' ')) data.splice(0, 1);
			var firstWord:String = words[0];
			name = data[0] as String;
			name = name.substr(0, name.length - 1);
			
			switch (firstWord)
			{
				case MTL_FILE:
					log("mtl file : " + data[0]);
					_mtlFile = data[0];
				break;
				case OBJECT_NAME:
					// let's finish the creation of the last polygon group for the precedent object (if exist)
					if (_currentGroup) {
						completePolygonGroup();
					}
					setNewObject(name);
				break;
				case VERTEX:
				if (!_currentObject) setNewObject("defaut_object"+_currentLineIndex);
					parseVertex(data);
				break;
				case NORMAL:
				if (!_currentObject) setNewObject("defaut_object"+_currentLineIndex);
					parseNormal(data);
				break;	
				case UV:
					if (!_currentObject) setNewObject("defaut_object"+_currentLineIndex);
					parseUv(data);
				break;				
					case GROUP_NAME:
					if (_currentGroup) {
						completePolygonGroup();
					}				
					setNewGroup(name);
				break;				
				case USE_MTL:
					var  mtlID:String = data[0];
					var found:Boolean = false;
					// we look for the mtl set corresponding 
					for each(var mtlSet:MMtlSet in _mtlSets) {
						if (mtlSet.id == mtlID) {
							_currentMtlSet = mtlSet;
							found = true;
						}
					}
					if(!found) 
						_currentMtlSet = _defaultMtlSet;
					writeMtl(_currentMtlSet);
				break;		
					case SMOOTH_SHADING:
					var pos:uint = _exportBA.position;
					_exportBA.position = _smoothPos;
					_exportBA.writeDouble(Number(data[0]));
					_exportBA.position = pos;
					_smoothValue = Number(data[0]);
				break;	
					case INDEX_DATA:
					// if indices are not part of an object and group we create them before
					if (!_currentGroup) setNewGroup("defaut_group" + _currentLineIndex);
					// if there's no mtl defined
					if (!_currentMtlSet) {
						_currentMtlSet = _defaultMtlSet;
						writeMtl(_currentMtlSet);
					}
					parseFaces(data);
				break;
			}
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function writeHeader():void 
		{
			_exportBA.writeUTFBytes("MDM");
			_exportBA.writeByte(1);
			_exportBA.writeByte(1);
			var empty:String = "";
			for (var i:int = 0; i < 97; i++) empty += "m";
			_exportBA.writeUTFBytes(empty);
		}
		//----------------------------------------------------------------------------------------------	
		// create a new object
		//----------------------------------------------------------------------------------------------		
		private function setNewObject(id:String):void {
			log("--------  object " + id);
			_currentObject = id;
			// writting object id
			_exportBA.writeUTFBytes("o");
			_exportBA.writeUTF(id);			
		}
		//----------------------------------------------------------------------------------------------	
		// create a new object
		//----------------------------------------------------------------------------------------------		
		private function setNewGroup(id:String):void {
			log("-- group " + id);
			_currentGroup = id;
			// a group exist and is not part of an object 
			if (!_currentObject) setNewObject("object_"+id);
			_exportBA.writeUTFBytes("g");
			// writting object id
			_exportBA.writeUTF(id);
			// writting default smooth shading
			_smoothPos = _exportBA.position;
			_smoothValue = 1;
			_exportBA.writeDouble(1);	
		}
		//----------------------------------------------------------------------------------------------	
		// write  a mtl set
		//----------------------------------------------------------------------------------------------		
		private function writeMtl(mtlSet:MMtlSet):void 
		{
			log("--  mtl : " + mtlSet.id );
			// write mtl set
			_exportBA.writeUTF(mtlSet.id);
			_exportBA.writeUnsignedInt(mtlSet.diffuseColor);
			_exportBA.writeUnsignedInt(mtlSet.ambiantColor);
			_exportBA.writeUnsignedInt(mtlSet.specularColor);
			_exportBA.writeDouble(mtlSet.specularCoef);	
			_exportBA.writeUnsignedInt(mtlSet.emissiveColor);
			_exportBA.writeDouble(mtlSet.dissolved);
			_exportBA.writeDouble(mtlSet.illumination);
			_exportBA.writeUTF(mtlSet.ambiantTexture);
			_exportBA.writeUTF(mtlSet.diffuseTexture);
			_exportBA.writeBoolean(mtlSet.isTexture);
		}
		//----------------------------------------------------------------------------------------------	
		//----------------------------------------------------------------------------------------------		
		private function completePolygonGroup():void {
			log("completing group");
			_currentGroup = null;
			_currentPolygoneNumber = 0;
			var pos:uint = _exportBA.position;
			_exportBA.position = _facesDataPos;
			trace("nb faces : " + _facesNumber);
			_exportBA.writeUnsignedInt(_facesNumber);
			_facesNumber = 0;
			_onFaces = false;
			_exportBA.position = pos;
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function parseVertex(data:Array):void {
			var pos:uint;
			if (!_onVertex) {
				if (_onNormal) {
					_onNormal = false;
					pos = _exportBA.position;
					_exportBA.position = _normalesDataPos;
					_exportBA.writeUnsignedInt(_normalesNumber);
					_exportBA.position = pos;
				}
				else if (_onUV) {
					_onUV = false;
					pos = _exportBA.position;
					_exportBA.position = _uvDataPos;
					_exportBA.writeUnsignedInt(_uvNumber);
					_exportBA.position = pos;
				}
				_onVertex = true;
				_exportBA.writeUTFBytes("v");
				_verticesDataPos = _exportBA.position;
				_verticesNumber = 0;
				_exportBA.writeUnsignedInt(0);
			}
			if (_vertexDataIsZxy) {
				_exportBA.writeFloat(Number(data[1]) * _scale);
				_exportBA.writeFloat(Number(data[2])*_scale);
				_exportBA.writeFloat(Number(data[0]) * _scale);
				
			}
			else { // normal operation: x,y,z
				_exportBA.writeFloat(Number(data[0]) * _scale);
				_exportBA.writeFloat(Number(data[1])*_scale);
				_exportBA.writeFloat(Number(data[2]) * _scale);
			}
			_verticesNumber ++;
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function parseNormal(data:Array):void {
			var pos:uint;
			if (!_onNormal) {
				if (_onVertex) {
					_onVertex = false;
					pos = _exportBA.position;
					_exportBA.position = _verticesDataPos;
					_exportBA.writeUnsignedInt(_verticesNumber);
					_exportBA.position = pos;
					
				}
				else if (_onUV) {
					_onUV = false;
					pos = _exportBA.position;
					_exportBA.position = _uvDataPos;
					_exportBA.writeUnsignedInt(_uvNumber);
					_exportBA.position = pos;
				}				
				_onNormal = true;
				_exportBA.writeUTFBytes("n");
				_normalesDataPos = _exportBA.position;
				_normalesNumber = 0;
				_exportBA.writeUnsignedInt(0);
			}			
			var loop:uint = data.length;
			if (loop > 3) {
				log("MObjParser more than 3 normales !!!");
				loop = 3;
			}
			for (var i:uint = 0; i < loop; ++i){
				var element:String = data[i];
				if (element != null) // handle 3dsmax extra spaces
				_exportBA.writeFloat(Number(element));
			}
			_normalesNumber ++;
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function parseUv(data:Array):void {
			var pos:uint;
			if (!_onUV) {
				if (_onVertex) {
					_onVertex = false;
					pos = _exportBA.position;
					_exportBA.position = _verticesDataPos;
					_exportBA.writeUnsignedInt(_verticesNumber);
					_exportBA.position = pos;
				}
				else if (_onNormal) {
					_onNormal = false;
					pos = _exportBA.position;
					_exportBA.position = _normalesDataPos;
					_exportBA.writeUnsignedInt(_normalesNumber);
					_exportBA.position = pos;
					
				}				
				_onUV = true;
				_exportBA.writeUTFBytes("t");
				_uvDataPos = _exportBA.position;
				_uvNumber = 0;
				_exportBA.writeUnsignedInt(0);
			}				
				var element:String = data[0];
				if(_mirrorUV){
				//OBJ has a flipped U axis for texture coordinates, so it needs to get flipped back to normal:
					_exportBA.writeFloat(Number(1 - Number(element)));
				}
				else{
					_exportBA.writeFloat(Number(element));
				}
				
				element = data[1];
				if(_mirrorUV){
				//OBJ has a flipped V axis for texture coordinates, so it needs to get flipped back to normal:
					_exportBA.writeFloat(Number(1 - Number(element)));
				}
				else{
					_exportBA.writeFloat(Number(element));
				}
				_uvNumber++;
		}
		//----------------------------------------------------------------------------------------------
		// explode the face into triangles 
		//----------------------------------------------------------------------------------------------		
		private function parseFaces(data:Array):void {
			var nbVertices:int = data.length;
			var pos:uint;
			if (_onVertex) {
				_onVertex = false;
				pos = _exportBA.position;
				_exportBA.position = _verticesDataPos;
				_exportBA.writeUnsignedInt(_verticesNumber);
				_exportBA.position = pos;
				
			}			
			else if (_onNormal) {
				_onNormal = false;
				pos = _exportBA.position;
				_exportBA.position = _normalesDataPos;
				_exportBA.writeUnsignedInt(_normalesNumber);
				_exportBA.position = pos;
				
			}
			else if (_onUV) {
				_onUV = false;
				pos = _exportBA.position;
				_exportBA.position = _uvDataPos;
				_exportBA.writeUnsignedInt(_uvNumber);
				_exportBA.position = pos;
			}
			if(!_onFaces){	
				_facesNumber = 0;
				_onFaces = true;
				_exportBA.writeUTFBytes("f");
				_facesDataPos = _exportBA.position;
				_exportBA.writeUnsignedInt(0);				
			}
			var triplet:Vector.<String> = new Vector.<String>(3, true);
			// Process elements turning faces into triangle fan.
			for (var i:int = 1; i < nbVertices-1; ++i) {
				triplet[0] = data[i];
				triplet[1] = data[0];
				triplet[2] = data[i + 1];
				parseTriplet(triplet);
			}
		}
		//----------------------------------------------------------------------------------------------	
		// for each triangle pass datas to the writer wich will add them in a byteArray
		//----------------------------------------------------------------------------------------------		
		private function parseTriplet(data:Vector.<String>):void {
			var triplet:String;
			var subdata:Array;
			var vertexIndex:uint;
			var normalIndex:uint;
			var uvIndex:uint;
			var index:uint;
			var loop:int = 3;
			var i:int;
			
			// loop through each element and grab values stored earlier
			// elements come as vertexIndex/normalIndex
			for(i = 0; i < loop; ++i){
				triplet = data[i];
				subdata = triplet.split(SLASH);
				vertexIndex = uint(subdata[0]) - 1;
				uvIndex = uint(subdata[1]) - 1;
				normalIndex = uint(subdata[2]) - 1;
				
				// sanity check
				if(vertexIndex < 0) vertexIndex = 0;
				if (normalIndex < 0) normalIndex = 0;
				if (uvIndex < 0) uvIndex = 0;
				
				_exportBA.writeUnsignedInt(vertexIndex);
				_exportBA.writeUnsignedInt(uvIndex);
				_exportBA.writeUnsignedInt(normalIndex);
			}
			_currentPolygoneNumber ++;
			_facesNumber ++;			
			
		}		
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function log(txt:String):void 
		{
			trace(txt);
		}	
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		public function get exportBA():ByteArray 
		{
			return _exportBA;
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function clean():void {
		}
		//----------------------------------------------------------------------------------------------	
		/**
		* delete datas
		*/				
		//----------------------------------------------------------------------------------------------	
		public function destroy():void {
			_mtlFile = null; 
			_mtlParser.destroy();
			_mtlParser = null;
			_exportBA.length = 0;
			_exportBA = null;
			_lineSignal.removeAll();
			_lineSignal= null;
		}
		

	}

}
