/*Copyright (c) 2012 Moussa Dembélé (www.mousman.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package com.mousman.as3.agal.parser.objs 
{
	import com.mousman.as3.agal.parser.dto.MBuffersWriterDTO;
	import com.mousman.as3.agal.parser.shaders.ShaderSet;
	import com.mousman.as3.agal.parser.shaders.ShaderStore;
	import com.mousman.as3.agal.shader.ShaderType;
	import com.mousman.as3.structure.garbage.IDestroyable;
	import com.mousman.as3.utils.color.ColorUtil;
	import flash.display3D.Context3D;
	import flash.display3D.IndexBuffer3D;
	import flash.display3D.VertexBuffer3D;
	import flash.utils.ByteArray;
	import mx.core.ByteArrayAsset;
	/**
	 * import a mdm compact binary file and create a set of vertex buffers and index buffers 
	 * @author Moussa Dembélé (www.mousman.com) based on Christer Kaitila and Nathan Ostgard work
	 * 
	 */
	public class MCompactObjImportParser implements IDestroyable
	{
		// the context3D in wich we want to upload the buffers
		private var _context3D:Context3D;		
		
		// contains all the datas for the mesh
		/**
		 * @private
		 */		
		private var _mesh:MObjMesh;
		
		// contains colors that we can force in the construction of the mesh
		private var _writerParams:MBuffersWriterDTO;
		// contains all the shaders we want to use
		private var _shaderStore:ShaderStore;
		
		private var _scale:Number;
		private var _vertices:Vector.<Number>;
		private var _normals:Vector.<Number>;
		private var _uvs:Vector.<Number>;
		
		private var _currentObject:MObject3D;
		private var _currentGroup:MPolygoneGroup3D;
		// contains all the polygons group for the current object
		private var _mGroups:Vector.<MPolygoneGroup3D>;
		
		private var _currentShaderSet:ShaderSet;
		private var _defaultShaderSet:ShaderSet;
		private var _defaultMtlSet:MMtlSet;		
		
		private var _compactBA:ByteArrayAsset;
		
		/**
		* Constructor. Creates a new MObjParser instance.
		*
		* <p>After you create the MObjParser instance, call the parse method to create the mesh datas.</p>
		* 
		* @param context : Context3D.<p>Used to create and upload VertexBuffers and IndexBuffers</p>
		* @param shaderStore : ShaderStore.<p>Contains all the shaders used for the mesh</p>
		*/		
		public function MCompactObjImportParser(context:Context3D, scale:Number, shaderStore:ShaderStore) {
			_context3D = context;
			_scale = scale;
			_shaderStore = shaderStore;
		}
		//----------------------------------------------------------------------------------------------
		/**
		* Parse the obj and mtl file and create the mesh datas
		*
		* 
		* @param meshData : ByteArrayAsset.<p>Contains the obj and mtl datas in a binary format</p>
		* @param writerParams : MBuffersWriterDTO <p>used to force color on the mesh.</p>
		* 
		* @return the datas of the mesh
		*/			
		//----------------------------------------------------------------------------------------------		
		public function parse(meshData:ByteArrayAsset, writerParams:MBuffersWriterDTO = null): MObjMesh {
			_mesh = new MObjMesh();
			if(_writerParams == null)
				_writerParams = new MBuffersWriterDTO();
				
			_vertices = new Vector.<Number>();
			_normals = new Vector.<Number>();		
			_uvs = new Vector.<Number>();
			
			// read and explode the datas
			_compactBA = meshData;
			parseObj();
			
			return _mesh;
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function parseObj():void {
			var shader:ShaderSet;	
			// initialisation of the writers, passing them vertices, normals and _uv references
			for each(shader in _shaderStore.shaderSetArray) {
				shader.writer.init(_vertices, _normals,_uvs);
			}			
			readHeader();
			var ret:String = _compactBA.readUTFBytes(1); 
			while (ret =="o") { ret = readObj() };
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function readHeader():void 
		{
			var signature:String = _compactBA.readUTFBytes(3);
			var blank:String = _compactBA.readUTFBytes(99);
		}		
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function readObj():String 
		{
			var id:String = _compactBA.readUTF();
			trace(" object " + id);
			_currentObject = new MObject3D(id);
			_mesh.add(_currentObject );
			
			_mGroups = new Vector.<MPolygoneGroup3D>();
			_currentObject.polygonsGroups = _mGroups;				
			
			var ret:String = "";
			
			var type:String = _compactBA.readUTFBytes(1);
			while (type != "o") { 
				readObjContent(type);
				// end of file
				if (_compactBA.position == _compactBA.length) {
					return null;
				}
				type = _compactBA.readUTFBytes(1);
			}
			
			return type;
			
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function readObjContent(type:String):void 
		{
			trace("type : " + type);
			switch (type) {
				case "v":
					readVerticesList();
					break;
				case "n":
					readNormalesList();
					break;
				case "t" :
					readUVsList();
					break;
				case "g" :
					readGroup();
					break;
				default :
					trace("MCompactObjImportParser : error reading object , unknown type : " + type);
					break;
			}
			
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function readVerticesList():void 
		{
			var val:Number; 
			var nbVertices:uint = _compactBA.readUnsignedInt();
			for (var i:int ; i < nbVertices; i++) {
				//x 
				val = Number(_compactBA.readFloat()) * _scale;
				_vertices.push(val);			
				//y
				val = Number(_compactBA.readFloat()) * _scale;
				_vertices.push(val);			
				//xz
				val = Number(_compactBA.readFloat()) * _scale;
				_vertices.push(val);			
			}
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function readNormalesList():void 
		{
			var val:Number; 
			var nbVertices:uint = _compactBA.readUnsignedInt();
			for (var i:int ; i < nbVertices; i++) {
				//x 
				val = Number(_compactBA.readFloat()) * _scale;
				_normals.push(val);			
				//y
				val = Number(_compactBA.readFloat()) * _scale;
				_normals.push(val);			
				//xz
				val = Number(_compactBA.readFloat()) * _scale;
				_normals.push(val);			
			}
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function readUVsList():void 
		{
			var val:Number; 
			var nbVertices:uint = _compactBA.readUnsignedInt();
			for (var i:int ; i < nbVertices; i++) {
				//u 
				val = Number(_compactBA.readFloat()) * _scale;
				_uvs.push(val);			
				//v
				val = Number(_compactBA.readFloat()) * _scale;
				_uvs.push(val);			
			}	
		}	
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function readGroup():void 
		{
			var id:String = _compactBA.readUTF();
			trace("-- group " + id);
			_currentGroup = new MPolygoneGroup3D(id);
			_mGroups.push(_currentGroup);
			_currentGroup.smoothShading = Number(_compactBA.readDouble());
			readMtlSet();
			
			// we look for a shader associated with the group 
			if(_shaderStore.getShaderById(id)){
				_currentShaderSet = _shaderStore.getShaderById(id); 
			}
			// else we use the one define for the object if it exists 
			else if(_shaderStore.getShaderById(_currentObject.id)){
				_currentShaderSet = _shaderStore.getShaderById(_currentObject.id); 
			}			
			// no shaderSet has been defined so we use the default ones
			else  {
				if (_currentGroup.mtlSet.isTexture) {
					_currentShaderSet = _shaderStore.getShaderById(ShaderType.DEFAULT_TEXTURE_SHADER);
				}
				else {
					_currentShaderSet = _shaderStore.getShaderById(ShaderType.DEFAULT_SHADER);
				}
			}
			
			// set the diffuse color as vertex color for the group
			var color:Object = ColorUtil.hexArgb(_currentGroup.mtlSet.diffuseColor);
			var alpha:int = _currentGroup.mtlSet.dissolved*255;
			var tcolor:uint = ColorUtil.argbHex(alpha, color.bleu, color.vert, color.rouge);
			_currentShaderSet.writer.verticesColor = tcolor;			
			
			readFaces();
			
			_currentShaderSet.writer.complete();
			createBuffers();
			_currentGroup = null;
			_currentShaderSet.writer.reset();
			_currentShaderSet = null;			
			
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function readMtlSet():void 
		{
			var mtlId:String = _compactBA.readUTF();
			trace("---- mtl " + mtlId);
			var mtlSet:MMtlSet = new MMtlSet(mtlId);
			mtlSet.diffuseColor = _compactBA.readUnsignedInt();
			mtlSet.ambiantColor = _compactBA.readUnsignedInt();
			mtlSet.specularColor = _compactBA.readUnsignedInt();
			mtlSet.specularCoef= _compactBA.readDouble();
			mtlSet.emissiveColor = _compactBA.readUnsignedInt();
			mtlSet.dissolved = _compactBA.readDouble();
			mtlSet.illumination = _compactBA.readDouble();
			var aTexture:String = _compactBA.readUTF();
			if (aTexture !="")
				mtlSet.ambiantTexture = aTexture;
			var dTexture:String = _compactBA.readUTF();
			if (dTexture !="")
				mtlSet.diffuseTexture = dTexture;
			var isTexture:Boolean = _compactBA.readBoolean();			
			
			_currentGroup.mtlSet = mtlSet;
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function readFaces():void 
		{
			var type:String = _compactBA.readUTFBytes(1);
			if (type != "f") trace("Error reading faces");
			var nbFaces:uint = _compactBA.readUnsignedInt();
			trace("    nb faces : " +nbFaces);
			var indexTriplet:Vector.<uint> = new Vector.<uint>(3,true);
			var normalTriplet:Vector.<uint> = new Vector.<uint>(3,true);
			var uvTriplet:Vector.<uint> = new Vector.<uint>(3,true);
			var vertexIndex:uint;
			var normalIndex:uint;
			var uvIndex:uint;
			var index:uint;
			
			for (var i:int; i < nbFaces; i++) {
				for(var j:int = 0; j < 3; ++j){
					indexTriplet[j] = _compactBA.readUnsignedInt();
					uvTriplet[j] = _compactBA.readUnsignedInt();
					normalTriplet[j] = _compactBA.readUnsignedInt();
				}
			_currentShaderSet.writer.addTriangle(indexTriplet, normalTriplet, uvTriplet);				
			}
		}
		//----------------------------------------------------------------------------------------------		
		// create the VertexBuffer3D and IndexBuffer3D for the current group
		//----------------------------------------------------------------------------------------------		
		private function createBuffers():void {
			var vertexByteArray:ByteArray;
			var vertexBuffer:VertexBuffer3D;
			var indexBuffer:IndexBuffer3D;
			var rawIndexBuffer:Vector.<uint>;
			var pGroupBuffers:MPolygonGroupBuffers;
			var nbVertices:int;
			var i:int;
			var dataPerVertex:int = _currentShaderSet.shader.dataPerVertex;

			nbVertices = _currentShaderSet.writer.nbVertices;
			_currentGroup.nbVertices = nbVertices;
			vertexByteArray = _currentShaderSet.writer.bytesArray;
			
			var next:int = 0;
			var rest:int = 0;
			var uploaded:int = 0;

			rest = nbVertices - uploaded;
			while (rest) {
				if ( rest > 65535) rest = 65535;
				pGroupBuffers = new MPolygonGroupBuffers(_currentGroup.id+"_"+i);
				// vertices
				vertexBuffer = _context3D.createVertexBuffer(rest, dataPerVertex);
				trace("creating " + rest +" vertex and index");
				vertexBuffer.uploadFromByteArray(vertexByteArray, uploaded * dataPerVertex * 4, 0, rest);
				// index
				indexBuffer = _context3D.createIndexBuffer(rest);
				rawIndexBuffer = _currentShaderSet.writer.indices.slice(uploaded, uploaded + rest);
				indexBuffer.uploadFromVector(rawIndexBuffer, 0, rest);
				
				pGroupBuffers.vertexBuffer = vertexBuffer;
				pGroupBuffers.indexBuffer = indexBuffer;
				_currentGroup.addPolygonBuffers(pGroupBuffers);
				
				uploaded += rest;
				rest = nbVertices - uploaded;
				i++;
			}
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function clean():void {
		}
		//----------------------------------------------------------------------------------------------	
		/**
		* delete datas
		*/				
		//----------------------------------------------------------------------------------------------	
		public function destroy():void {
			_mesh.destroy();
			_mesh = null;
			_writerParams = null;
			_shaderStore.destroy();
			_shaderStore = null;
			_context3D = null;
		}
		//----------------------------------------------------------------------------------------------	
		/**
		* the object containing mesh datas
		*/			
		//----------------------------------------------------------------------------------------------	
		/**
		* the object containing mesh datas
		*/			
		public function get mesh():MObjMesh {return _mesh;}
	}

}
