/*Copyright (c) 2012 Moussa Dembélé (www.mousman.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package com.mousman.as3.agal.parser.objs 
{
	import com.mousman.as3.agal.parser.dto.MBuffersWriterDTO;
	import com.mousman.as3.structure.garbage.IDestroyable;
	import com.mousman.as3.utils.color.ColorUtil;
	/**
	 * Parse a MTL file and create a Vector of MTLSet
	 * @author Moussa Dembélé (www.mousman.com)
	 */
	public class MMtlParser implements IDestroyable
	{
		private const LINE_FEED:String = String.fromCharCode(10);
		private const SPACE:String = String.fromCharCode(32);
		private const SLASH:String = "/";
		private const NEWMTL:String = "newmtl";
		private const DIFFUSE_COLOR:String = "Kd";
		private const AMBIANT_COLOR:String = "Ka";
		private const SPECULAR_COLOR:String = "Ks";
		private const SPECULAR_COEF:String = "Ns";
		private const EMISSIVE_COLOR:String = "Ke";
		private const DISSOLVED_A:String = "d";
		private const DISSOLVED_B:String = "Tr";
		private const ILLUMINATION:String = "illum";
		private const AMBIANT_TEXTURE:String = "map_Ka";
		private const DIFFUSE_TEXTURE:String = "map_Kd";
		
		private var _mtlFile:String;		
		private var _mtlSet:MMtlSet;	
		/**
		 * @private
		 */				
		private var _mtlSetVector:Vector.<MMtlSet>;		
		
		private var _isATexture:Boolean = false;
		private var _isDTexture:Boolean = false;		
		private var _isDColor:Boolean = false;		
		private var _isEColor:Boolean = false;		
		private var _isAColor:Boolean = false;		
		private var _isSColor:Boolean = false;		
		private var _isSCoef:Boolean = false;		
		
		private var _writerParams:MBuffersWriterDTO;		
		
		/**
		* Constructor. Creates a new MMtlParser instance.
		*
		* @param file : String.<p>the mtl file as a String</p>
		*/			
		public function MMtlParser(file:String) 
		{
			_mtlFile =  file;
		}
		//----------------------------------------------------------------------------------------------		
		/**
		* parse the mtl file
		*
		* @param writerParams : MBuffersWriterDTO.<p>used to force colors</p>
		*/			
		//----------------------------------------------------------------------------------------------		
		public function parse(writerParams:MBuffersWriterDTO=null):void {
			var lines:Array = _mtlFile.split(LINE_FEED);
			var loop:uint = lines.length;			
			
			if(_writerParams)
				_writerParams = writerParams;
				
			_mtlSetVector = new Vector.<MMtlSet>();
			
			for (var i:uint = 0; i < loop; ++i) {
				parseLine(lines[i]);
			}
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function parseLine(line:String):void {
			// Split line into words.
			var words:Array = line.split(SPACE);
			// Prepare the data of the line.
			if (words.length > 0) var data:Array = words.slice(1);
			else return;
			// Check first word and delegate remainder to proper parser.
			var firstWord:String = words[0];
			while ((data[0] == '') || (data[0] == ' ')) data.splice(0, 1);
			
			switch (firstWord)
			{
				case NEWMTL:
					// we check for completing the previosus mtl data
					if (_mtlSet ) {
						if (!_isDColor) _mtlSet.diffuseColor = 0xff000000;
						if (!_isAColor) _mtlSet.ambiantColor = _mtlSet.diffuseColor;
						if (!_isSColor) _mtlSet.specularColor = _mtlSet.diffuseColor;
						if (!_isSCoef) _mtlSet.specularCoef = 1;
						if (!_isEColor) _mtlSet.emissiveColor = 0;
						
						// if some colors are forced
						if(_writerParams){
							if(_writerParams.forceAmbiantColor ) _mtlSet.ambiantColor = _writerParams.ambiantColor;
							if (_writerParams.forceDiffuseColor ) _mtlSet.diffuseColor = _writerParams.diffuseColor;
							if(_writerParams.forceEmissiveColor ) _mtlSet.emissiveColor = _writerParams.emissiveColor;
							if(_writerParams.forceSpecularColor) _mtlSet.specularColor = _writerParams.specularColor;
							if (_writerParams.forceSpecularColor) _mtlSet.specularCoef = _writerParams.specularCoef;
						}
					}
					
					_mtlSet = new MMtlSet(data[0]);
					_mtlSetVector.push(_mtlSet);
					_isATexture = false;
					_isDTexture = false;
					_isAColor = false;
					_isSColor = false;
					_isSCoef = false;
					//trace("creating new mtlset : " + _mtlSet.id);
				break;				
				case DIFFUSE_COLOR:
					_mtlSet.diffuseColor = parseColor(data);
					_isDColor = true;
					//trace("    golbal color: " + _mtlSet.globalColor.toString(16) );
				break;
				case AMBIANT_COLOR:
					_mtlSet.ambiantColor = parseColor(data);
					_isAColor = true;
					//trace("    ambiant color: " + _mtlSet.globalColor.toString(16) );
				break;
				case SPECULAR_COLOR:
					_mtlSet.specularColor = parseColor(data);
					_isSColor = true;
					//trace("    specular color: " + _mtlSet.globalColor.toString(16) );
				break;
				case SPECULAR_COEF:
					_mtlSet.specularCoef = Number(data[0]);
					_isSCoef = true;
					//trace("    specularCoef: " + _mtlSet.specularCoef );
				break;
				case EMISSIVE_COLOR:
					_mtlSet.emissiveColor = Number(data[0]);
					_isEColor = true;
					//trace("    specularCoef: " + _mtlSet.specularCoef );
				break;
				case DISSOLVED_A:
				case DISSOLVED_B:
					_mtlSet.dissolved = Number(data[0]);
					//trace("    dissolved: " + _mtlSet.dissolved);
				break;
				case ILLUMINATION:
					_mtlSet.illumination = int(data[0]);
					//trace("    illumination: " + _mtlSet.illumination );
				break;
				case AMBIANT_TEXTURE:
					_mtlSet.ambiantTexture = String(data[0]).split(".")[0];
					_isATexture = true;
					//trace("    ambiantTexture: " + _mtlSet.ambiantTexture );
				break;
			case DIFFUSE_TEXTURE:
					_mtlSet.diffuseTexture = String(data[0]).split(".")[0];
					_isDTexture = true;
					//trace("    diffuseTexture: " + _mtlSet.diffuseTexture);
				break;
			}
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function parseColor(data:Array):int 
		{
			var color:int;
			var red:int = Number(data[0]) * 255;
			var green:int = Number(data[1]) * 255;
			var blue:int = Number(data[2]) * 255;
			color = ColorUtil.argbHex(0xff, red, green, blue);
			return color;
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		/**
		* Vector of MtlSet
		*/ 				
		public function get mtlSetVector():Vector.<MMtlSet> {return _mtlSetVector;}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function destroy():void {
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
	}

}