/*Copyright (c) 2012 Moussa Dembélé (www.mousman.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package com.mousman.as3.agal.parser.objs 
{
	import com.mousman.as3.agal.parser.dto.MBuffersWriterDTO;
	import com.mousman.as3.agal.parser.shaders.ShaderSet;
	import com.mousman.as3.agal.parser.shaders.ShaderStore;
	import com.mousman.as3.agal.shader.ShaderType;
	import com.mousman.as3.structure.garbage.IDestroyable;
	import com.mousman.as3.utils.color.ColorUtil;
	import flash.display3D.Context3D;
	import flash.display3D.IndexBuffer3D;
	import flash.display3D.VertexBuffer3D;
	import flash.utils.ByteArray;
	/**
	 * create a set of vertex buffers and index buffers based on a .obj and .mtl file
	 * @author Moussa Dembélé (www.mousman.com) based on Nathan Ostgard work
	 * 
	 */
	public class MObjParser implements IDestroyable
	{
		// constants used in parsing OBJ data
		private const LINE_FEED:String = String.fromCharCode(10);
		private const SPACE:String = String.fromCharCode(32);
		private const SLASH:String = "/";
		private const VERTEX:String = "v";
		private const NORMAL:String = "vn";
		private const UV:String = "vt";
		private const INDEX_DATA:String = "f";	
		private const MTL_FILE:String = "mtllib";	
		private const USE_MTL:String = "usemtl";	
		private const OBJECT_NAME:String = "o";	
		private const GROUP_NAME:String = "g";	
		private const SMOOTH_SHADING:String = "s";
		
		private var _mtlFile:String;		
		// older versions of 3dsmax use an invalid vertex order:
		private var _vertexDataIsZxy:Boolean = false;
		
		private var _scale:Number;
		private var _vertices:Vector.<Number>;
		private var _normals:Vector.<Number>;
		private var _uvs:Vector.<Number>;

		// the context3D in wich we want to upload the buffers
		private var _context3D:Context3D;		
		
		// the parser for the mtl datas
		private var _mtlParser:MMtlParser;
		// contains all the mtl datas parsed for the mtl file
		private var _mtlSets:Vector.<MMtlSet>;
		
		private var _currentObject:MObject3D;
		private var _currentGroup:MPolygoneGroup3D;
		// contains all the polygons group for the current object
		private var _mGroups:Vector.<MPolygoneGroup3D>;
		
		private var _currentShaderSet:ShaderSet;
		private var _defaultShaderSet:ShaderSet;
		private var _defaultMtlSet:MMtlSet;
		
		// contains all the datas for the mesh
		/**
		 * @private
		 */
		private var _mesh:MObjMesh;
		
		// contains colors that we can force in the construction of the mesh
		private var _writerParams:MBuffersWriterDTO;
		// contains all the shaders we want to use
		private var _shaderStore:ShaderStore;
		// some exporters mirror the UV texture coordinates
		/**
		 * @private
		 */		
		private var _mirrorUV:Boolean = true;
		
		/**
		* Constructor. Creates a new MObjParser instance.
		*
		* <p>After you create the MObjParser instance, call the parse method to create the mesh datas.</p>
		* 
		* @param context : Context3D.<p>Used to create and upload VertexBuffers and IndexBuffers</p>
		* @param scale : Number.<p>Used to resize the object</p>
		* @param shaderStore : ShaderStore.<p>Contains all the shaders used for the mesh</p>
		* @param dataIsZxy : Boolean.<p>Older versions of 3dsmax use an invalid vertex order</p>
		* @param mirrorUV : Boolean.<p>Some exporters mirror the UV texture coordinates</p>
		*/		
		public function MObjParser(context:Context3D, scale:Number, shaderStore:ShaderStore, dataIsZxy:Boolean = false,mirrorUV:Boolean=true) {
			// older versions of 3dsmax use an invalid vertex order:
			_vertexDataIsZxy = dataIsZxy;
			// some exporters mirror the UV texture coordinates
			_mirrorUV = mirrorUV;
			_scale = scale;
			_context3D = context;
			_shaderStore = shaderStore;
		}
		//----------------------------------------------------------------------------------------------
		/**
		* Parse the obj and mtl file and create the mesh datas
		*
		* 
		* @param objfile : String.<p>Contains the obj file datas</p>
		* @param mtlfile : String.<p>Contains the mtl file datas</p>
		* @param writerParams : MBuffersWriterDTO<p>used to force color on the mesh.</p>
		* 
		* @return the datas of the mesh
		*/			
		//----------------------------------------------------------------------------------------------		
		public function parse(objfile:String, mtlfile:String = null, writerParams:MBuffersWriterDTO = null): MObjMesh {
			clean();
			
			_mesh = new MObjMesh();
			if(writerParams == null) 
				_writerParams = new MBuffersWriterDTO();
			// read and explode the mtl datas
			if(mtlfile)
				parseMtl(mtlfile);
			_defaultMtlSet = new MMtlSet("default");
			_defaultMtlSet.ambiantColor = 0xcccccc;
			_defaultMtlSet.diffuseColor = 0xcccccc;
			_defaultMtlSet.dissolved= 1;
			_defaultMtlSet.emissiveColor = 0x0;
			_defaultMtlSet.specularCoef = 100;
			_defaultMtlSet.specularColor= 0xffffff;
			// read and explode the obj datas
			parseObj(objfile);
			
			return _mesh;
		}
		//----------------------------------------------------------------------------------------------
		// parse the mtl file and create a set of MMtlSet
		//----------------------------------------------------------------------------------------------
		private function parseMtl(mtlfile:String):void {
			_mtlParser = new MMtlParser(mtlfile);
			_mtlParser.parse(_writerParams);
			_mtlSets = _mtlParser.mtlSetVector;
		}	
		//----------------------------------------------------------------------------------------------
		// parse the obj file 
		//----------------------------------------------------------------------------------------------
		private function parseObj(objfile:String):void {
			
			var lines:Array;

			lines = objfile.split(LINE_FEED);
			var line:String;
			var loop:uint = lines.length;		
			var id:String;
			var shader:ShaderSet;	
			
			_vertices = new Vector.<Number>();
			_normals = new Vector.<Number>();		
			_uvs = new Vector.<Number>();
			
			// initialisation of the writers, passing them vertices, normals and _uv references
			for each(shader in _shaderStore.shaderSetArray) {
				shader.writer.init(_vertices, _normals,_uvs);
			}
			
			// parsing each line of the obj file
			for (var i:int = 0; i < loop; ++i) {
				line = lines[i];
				line.replace(/^\s*|\s*$/g, '');
				parseLine(line);
			}
			 //last polygon group
			completePolygonGroup();
		}
		//----------------------------------------------------------------------------------------------	
		//----------------------------------------------------------------------------------------------		
		private function parseLine(line:String):void {
			var name:String;
			// Split line into words.
			var words:Array = line.split(SPACE);
			// Blank line or comment, ignore it
			if (words.length ==0 || line === '' || line.charAt(0) === '#') {
				return;
			}
			// Prepare the data of the line.
			var data:Array = words.slice(1);
			
			// Check first word and delegate remainder to proper parser.
			while ((data[0] == '') || (data[0] == ' ')) data.splice(0, 1);
			var firstWord:String = words[0];
			name = data[0] as String;
			name = name.substr(0, name.length - 1);
			
			switch (firstWord)
			{
				case MTL_FILE:
					log("mtl file : " + data[0]);
					_mtlFile = data[0];
				break;
				case OBJECT_NAME:
					// let's finish the creation of the last polygon group for the precedent object (if exists)
					if (_currentGroup) {
						completePolygonGroup();
					}
					setNewObject(name);
					// if we have a shader defined for this object
					if (_shaderStore.getShaderById(name)) {
						_currentShaderSet = _shaderStore.getShaderById(name);
					}
				break;
				case VERTEX:
					parseVertex(data);
				break;
				case NORMAL:
					parseNormal(data);
				break;	
				case UV:
					parseUv(data);
				break;				
				case GROUP_NAME:
					log("   group : " + name);
					// let's finish the creation of the precedent polygon group
					if (_currentGroup) {
						completePolygonGroup();
					}
					_currentGroup = new MPolygoneGroup3D(name);
					// a group exist and is not part of an object 
					if (!_currentObject) setNewObject(name);
					_mGroups.push(_currentGroup);
					// we look for a shader associated with the group 
					if(_shaderStore.getShaderById(name)){
						_currentShaderSet = _shaderStore.getShaderById(name); 
					}
					// else we use the one define for the object if it exists 
					else if(_shaderStore.getShaderById(_currentObject.id)){
						_currentShaderSet = _shaderStore.getShaderById(_currentObject.id); 
					}
				break;				
				case USE_MTL:
					log("   mtl : " + data[0]);
					var  mtlID:String = data[0];
					var found:Boolean = false;
					// we look for the mtl set corresponding 
					for each(var mtlSet:MMtlSet in _mtlSets) {
						if (mtlSet.id == mtlID) {
							_currentGroup.mtlSet = mtlSet;
							found = true;
						}
					}
					if(!found) 
						_currentGroup.mtlSet = _defaultMtlSet;
					
					// no shaderSet has been defined so we use the default ones
					if (!_currentShaderSet) {
						if (_currentGroup.mtlSet.isTexture) {
							_currentShaderSet = _shaderStore.getShaderById(ShaderType.DEFAULT_TEXTURE_SHADER);
						}
						else {
							_currentShaderSet = _shaderStore.getShaderById(ShaderType.DEFAULT_SHADER);
						}
					}
					
					// set the diffuse color as vertex color for the group
					var color:Object = ColorUtil.hexArgb(_currentGroup.mtlSet.diffuseColor);
					var alpha:int = _currentGroup.mtlSet.dissolved*255;
					var tcolor:uint = ColorUtil.argbHex(alpha, color.bleu, color.vert, color.rouge);
					_currentShaderSet.writer.verticesColor = tcolor;
				break;		
				case SMOOTH_SHADING:
					_currentGroup.smoothShading = Number(data[0]);
				break;	
				case INDEX_DATA:
					parseFaces(data);
				break;
			}
		}
		//----------------------------------------------------------------------------------------------	
		// create a new object
		//----------------------------------------------------------------------------------------------		
		private function setNewObject(data:String):void {
			_currentObject = new MObject3D(data);
			_mesh.add(_currentObject);
			_mGroups = new Vector.<MPolygoneGroup3D>();
			_currentObject.polygonsGroups = _mGroups;		
		}
		//----------------------------------------------------------------------------------------------	
		//----------------------------------------------------------------------------------------------		
		private function completePolygonGroup():void {
			_currentShaderSet.writer.complete();
			createBuffers();
			_currentGroup = null;
			_currentShaderSet.writer.reset();
			_currentShaderSet = null;
		}
		//----------------------------------------------------------------------------------------------
		// explode the face into triangles 
		//----------------------------------------------------------------------------------------------		
		private function parseFaces(data:Array):void{
			var nbVertices:int = data.length;
			var triplet:Vector.<String> = new Vector.<String>(3, true);
			// Process elements turning faces into triangle fan.
			for (var i:int = 1; i < nbVertices-1; ++i) {
				triplet[0] = data[i];
				triplet[1] = data[0];
				triplet[2] = data[i + 1];
				parseTriplet(triplet);
			}
		}
		//----------------------------------------------------------------------------------------------	
		// for each triangle pass datas to the writer wich will add them in a byteArray
		//----------------------------------------------------------------------------------------------		
		private function parseTriplet(data:Vector.<String>):void {
			var triplet:String;
			var indexTriplet:Vector.<uint> = new Vector.<uint>(3,true);
			var normalTriplet:Vector.<uint> = new Vector.<uint>(3,true);
			var uvTriplet:Vector.<uint> = new Vector.<uint>(3,true);
			var subdata:Array;
			var vertexIndex:uint;
			var normalIndex:uint;
			var uvIndex:uint;
			var index:uint;
			var loop:int = 3;
			var i:int;
			
			// loop through each element 
			// elements come as vertexIndex/uvIndex/normalIndex
			for(i = 0; i < loop; ++i){
				triplet = data[i];
				subdata = triplet.split(SLASH);
				vertexIndex = uint(subdata[0]) - 1;
				uvIndex = uint(subdata[1]) - 1;
				normalIndex = uint(subdata[2]) - 1;
				
				// sanity check
				if(vertexIndex < 0) vertexIndex = 0;
				if (normalIndex < 0) normalIndex = 0;
				if (uvIndex < 0) uvIndex = 0;
				
				indexTriplet[i] = vertexIndex;
				normalTriplet[i] = normalIndex;
				uvTriplet[i] = uvIndex;
			}

			_currentShaderSet.writer.addTriangle(indexTriplet, normalTriplet, uvTriplet);
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function parseVertex(data:Array):void {
			if (_vertexDataIsZxy){
				_vertices.push(Number(data[1])*_scale);
				_vertices.push(Number(data[2])*_scale);
				_vertices.push(Number(data[0])*_scale);
			}
			else { // normal operation: x,y,z
				_vertices.push(Number(data[0])*_scale);
				_vertices.push(Number(data[1])*_scale);
				_vertices.push(Number(data[2])*_scale);
			}
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function parseNormal(data:Array):void{
			var loop:uint = data.length;
			if (loop > 3) {
				log("MObjParser more than 3 normales !!!");
				loop = 3;
			}
			for (var i:uint = 0; i < loop; ++i){
				var element:String = data[i];
				if (element != null) // handle 3dsmax extra spaces
				_normals.push(Number(element));
			}
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function parseUv(data:Array):void{
				var element:String = data[0];
				if(_mirrorUV)
				//OBJ has a flipped U axis for texture coordinates, so it needs to get flipped back to normal:
					_uvs.push(1 - Number(element));
				else
					_uvs.push(Number(element));
				
				element = data[1];
				if(_mirrorUV)
				//OBJ has a flipped V axis for texture coordinates, so it needs to get flipped back to normal:
					_uvs.push(1 - Number(element));
				else
					_uvs.push(Number(element));
		}
		//----------------------------------------------------------------------------------------------		
		// create the VertexBuffer3D and IndexBuffer3D for the current group
		//----------------------------------------------------------------------------------------------		
		private function createBuffers():void {
			var vertexByteArray:ByteArray;
			var vertexBuffer:VertexBuffer3D;
			var indexBuffer:IndexBuffer3D;
			var rawIndexBuffer:Vector.<uint>;
			var pGroupBuffers:MPolygonGroupBuffers;
			var nbVertices:int;
			var i:int;
			var dataPerVertex:int = _currentShaderSet.shader.dataPerVertex;

			nbVertices = _currentShaderSet.writer.nbVertices;
			_currentGroup.nbVertices = nbVertices;
			vertexByteArray = _currentShaderSet.writer.bytesArray;
			
			var next:int = 0;
			var rest:int = 0;
			var uploaded:int = 0;

			rest = nbVertices - uploaded;
			while (rest) {
				if ( rest > 65535) rest = 65535;
				pGroupBuffers = new MPolygonGroupBuffers(_currentGroup.id+"_"+i);
				// vertices
				vertexBuffer = _context3D.createVertexBuffer(rest, dataPerVertex);
				//log("creating " + rest +" vertex and index");
				vertexBuffer.uploadFromByteArray(vertexByteArray, uploaded * dataPerVertex * 4, 0, rest);
				// index
				indexBuffer = _context3D.createIndexBuffer(rest);
				rawIndexBuffer = _currentShaderSet.writer.indices.slice(uploaded, uploaded + rest);
				indexBuffer.uploadFromVector(rawIndexBuffer, 0, rest);
				
				pGroupBuffers.vertexBuffer = vertexBuffer;
				pGroupBuffers.indexBuffer = indexBuffer;
				_currentGroup.addPolygonBuffers(pGroupBuffers);
				
				uploaded += rest;
				rest = nbVertices - uploaded;
				i++;
			}
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function clean():void {
		}
		//----------------------------------------------------------------------------------------------	
		/**
		* delete datas
		*/				
		//----------------------------------------------------------------------------------------------	
		public function destroy():void {
			_mtlFile = null; 
			_vertices.length = 0;
			_vertices = null;
			_normals.length = 0;
			_normals = null;
			_uvs.length = 0;
			_uvs = null;
			_mtlParser.destroy();
			_mtlParser = null;
			_mtlSets.length = 0;
			_mtlSets = null;
			_currentObject = null;
			_currentGroup = null;
			_mGroups.length = 0;
			_mGroups = null;
			_currentShaderSet = null;
			_defaultShaderSet = null;
			_mesh.destroy();
			_mesh = null;
			_writerParams = null;
			_shaderStore.destroy();
			_shaderStore = null;
			_context3D = null;
		}
		//----------------------------------------------------------------------------------------------	
		/**
		* the object containing mesh datas
		*/			
		//----------------------------------------------------------------------------------------------	
		public function get mesh():MObjMesh {return _mesh;}
		//----------------------------------------------------------------------------------------------	
		/**
		* some exporters mirror the UV texture coordinates
		* @default true
		*/ 			
		//----------------------------------------------------------------------------------------------	
		public function set mirrorUV(value:Boolean):void { _mirrorUV = value; }
		//----------------------------------------------------------------------------------------------	
		//----------------------------------------------------------------------------------------------	
		public function log(txt:String):void {
			//trace(txt);
		}
	}

}
